#include <libregexis024fa/graph_to_bytecode/writing_commands.h>
#include <libregexis024vm/vm_opcodes.h>
#include <assert.h>

void cmd_JUMP(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, bookmark_id_t dest) {
    write_byte(result, regex024_opcodes::JUMP);
    bookmark_manager.write_unresolved_reference(result, dest);
}

constexpr regex024_opcode cmp_EQUAL[4] = {regex024_opcodes::JCEQUAL_B, regex024_opcodes::JCEQUAL_W,
    regex024_opcodes::JCEQUAL_DW, regex024_opcodes::JCEQUAL_QW};
constexpr regex024_opcode cmp_LESS[4] = {regex024_opcodes::JCLESS_B, regex024_opcodes::JCLESS_W,
    regex024_opcodes::JCLESS_DW, regex024_opcodes::JCLESS_QW};
constexpr regex024_opcode cmp_GRTR[4] = {regex024_opcodes::JCGRTR_B, regex024_opcodes::JCGRTR_W,
    regex024_opcodes::JCGRTR_DW, regex024_opcodes::JCGRTR_QW};


void cmd_JC(const regex024_opcode cmpT[4],
    std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest)
{
    if (val <= UINT8_MAX) {
        write_byte(result, cmpT[0]);
        write_byte(result, static_cast<uint8_t>(val));
    } else if (val <= UINT16_MAX) {
        write_byte(result, cmpT[1]);
        write_word(result, static_cast<uint16_t>(val));
    } else if (val <= UINT32_MAX) {
        write_byte(result, cmpT[2]);
        write_doubleword(result, static_cast<uint32_t>(val));
    } else {
        write_byte(result, cmpT[3]);
        write_quadword(result, val);
    }
    bookmark_manager.write_unresolved_reference(result, dest);
}


void cmd_JCEQUAL(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
    cmd_JC(cmp_EQUAL, result, bookmark_manager, val, dest);
}

void cmd_JCLESS(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
    cmd_JC(cmp_LESS, result, bookmark_manager, val, dest);
}

void cmd_JCGRTR(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
    cmd_JC(cmp_GRTR, result, bookmark_manager, val, dest);
}

void cmd_DIE(std::vector<uint8_t> &result) {
    write_byte(result, regex024_opcodes::DIE);
}

void cmd_MATCH(std::vector<uint8_t> &result) {
    write_byte(result, regex024_opcodes::MATCH);
}

void cmd_READ_first_ns(std::vector<uint8_t>& result, size_t slot) {
    assert(slot <= UINT32_MAX);
    write_byte(result, regex024_opcodes::READ);
    write_sslot_id(result, slot);
}

void cmd_FORK(std::vector<uint8_t> &result, explicit_bookmarks& bookmark_manager, size_t slot, bookmark_id_t dest) {
    assert(slot <= UINT32_MAX);
    write_byte(result, regex024_opcodes::FORK);
    write_sslot_id(result, slot);
    bookmark_manager.write_unresolved_reference(result, dest);
}

void cmd_READ_second_ns(std::vector<uint8_t>& result, std::vector<size_t>& belate_second_read_ns_slot_args) {
    write_byte(result, regex024_opcodes::READ);
    belate_second_read_ns_slot_args.push_back(result.size());
    write_sslot_id(result, 0);
}
