#include <libregexis024fa/tracking_fa_nodes.h>
#include <assert.h>

bool isImmMovOpcode(regex024_opcode inst) {
    return inst == regex024_opcodes::MOV_COLARR_IMM || inst == regex024_opcodes::MOV_SELARR_IMM;
}

bool isCurPosMovOpcode(regex024_opcode inst) {
    return inst == regex024_opcodes::MOV_COLARR_BTPOS || inst == regex024_opcodes::MOV_SELARR_CHPOS;
}

bool isColarrOpcode(regex024_opcode inst) {
    return inst == regex024_opcodes::MOV_COLARR_IMM || inst == regex024_opcodes::MOV_COLARR_BTPOS;
}

bool isSelarrOpcode(regex024_opcode inst) {
    return inst == regex024_opcodes::MOV_SELARR_IMM || inst == regex024_opcodes::MOV_SELARR_CHPOS;
}

bool isTrackingFaNode(const FA_Node *n) {
    return n->type == track_array_mov_imm || n->type == track_array_mov_halfinvariant;
}

TrackingOperationInFa::TrackingOperationInFa(regex024_opcode opcode, regex_tai_t key, uint64_t imm_value)
        : opcode(opcode), key(key), immValue(imm_value) {}

TrackingOperationInFa::TrackingOperationInFa(regex024_opcode opcode, regex_tai_t key)
    : opcode(opcode), key(key) {}

std::string TrackingOperationInFa::toString() const {
    switch (opcode){
        case regex024_opcodes::MOV_COLARR_IMM:
            return "colarr[" + std::to_string(key) + "] := " + std::to_string(immValue);
        case regex024_opcodes::MOV_SELARR_IMM:
            return "selarr[" + std::to_string(key) + "] := " + std::to_string(immValue);
        case regex024_opcodes::MOV_COLARR_BTPOS:
            return "colarr[" + std::to_string(key) + "] := cur byte position";
        case regex024_opcodes::MOV_SELARR_CHPOS:
            return "selarr[" + std::to_string(key) + "] := cur char position";
        default:
            return "wrong collection operation";
    }
}

FA_NodePathPart* convert_to_node(const TrackingOperationInFa& op, FA_Container& fa) {
    if (isImmMovOpcode(op.opcode)) {
        return fa.makeTrackArrayMovImm(op.opcode, op.key, op.immValue);
    }
    assert(isCurPosMovOpcode(op.opcode));
    return fa.makeTrackArrayMovHalfinvariant(op.opcode, op.key);

}

