#include <libregexis024tools/stringmatching.h>
#include <assert.h>
#include <stdexcept>
#include <libregexis024sol/expr_compiler.h>

using namespace regexis024;
using namespace std;

void test(const string& input, const string& pattern, const MatchInfo& right_answer) {
    MatchInfo given_answer;
    track_var_list retTrackVarList;
    string retStatus;
    int ret = matchStrToRegexp(input, pattern, given_answer, retTrackVarList, retStatus);
    if (ret < 0) {
        throw runtime_error("Test failed. matching. " + retStatus);
    }
    if (given_answer != right_answer) {
        throw runtime_error("Test failed.");
    }
    printf("Test passed\n");
}

int main() {
    test("C111111111111", "C\\>1*", MatchInfo({}, {}));
    // return 0;
    test("GET / HTTP/1.1\r\nHost: bibura sosat\r\nLos-es-raus: a\rfaafafdf\r\n\r\n",
        "!dfa;(GET|POST) / HTTP/(1.1|1.0|0.9)\r\n([\\u0021-\\u007E&^:]+:([\\u0000-\\u007F&^\r\n])*\r\n)*\r\n",
        MatchInfo());
    test("\r24234\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("\n3432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("3:::;;432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("3:::;;432   \r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("GET / HTTP/0.9\r\nHost: bibura sosat\r\nLos-es-raus: afaafafdf\r\n\r\n",
        "^(GET|POST\\>) / HTTP/(1.1|1.0|0.9)\r\n([\\u0021-\\u007E&^:]+:([\\u0000-\\u007F&^\r\n])*\r\n)*\r\n",
        MatchInfo({}, {}));
    // return 0;
    test("b", "#boba(b)", MatchInfo({{0, 0}, {1, 1}}, {}));
    test("abc", "!selarr{boba{ca}}^a#boba(b)c$", MatchInfo({{0, 1}, {1, 2}}, {1, 2}));
    for (int i = 0; i < 64; i++) {
        std::string T;
        T += ('a' + (i >> 3));
        T+= ('a' + (i % 8));
        test(T, "(((a|b)|(c|d))|((e|f)|(g|h)))!r{2}", MatchInfo({}, {}));
    }
    test("abba", "!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("abba", "!dfa;!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("abba", "!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("abba", "!dfa;!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("", "", MatchInfo({}, {}));
    test("a", "a", MatchInfo({}, {}));
    test("a3", "[abc]3", MatchInfo({}, {}));
    test("b3", "[abc]3", MatchInfo({}, {}));
    test("c3", "[abc]3", MatchInfo({}, {}));
    test("aa", "aa", MatchInfo({}, {}));
    test("aaaaa", "a*", MatchInfo({}, {}));
    test("bababbaa", "[ab]*", MatchInfo({}, {}));
    test("bababbaa", "!dfa;[ab]*", MatchInfo({}, {}));
    return 0;
}
