#ifndef LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H
#define LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H

#include <stdint.h>
#include <vector>
#include <libregexis024vm/vm_errno.h>
#include <libregexis024vm/vm_opcodes_types.h>

struct REGEX_IS024_CAEvent{
    regex_tai_t key;
    uint64_t value;
    bool operator==(const REGEX_IS024_CAEvent& other) const;
};

class REGEX_IS024_VirtualMachine{
public:
    REGEX_IS024_VirtualMachine(size_t programSize, const uint8_t *data, uint64_t caTreeLimit, uint16_t saLenLimit,
                               uint32_t readSsLimit, uint32_t forkSsLimit, uint64_t timeTickLimit);

    REGEX_IS024_VirtualMachine(const REGEX_IS024_VirtualMachine& ) = delete;
    REGEX_IS024_VirtualMachine& operator=(const REGEX_IS024_VirtualMachine&) = delete;

    regex024_error_code initialize();
    bool isInitialized();
    bool isUsable();
    virtual ~REGEX_IS024_VirtualMachine();
    regex_tai_t getSelectionArrayLength();
    bool isAllowMultistart();
    uint8_t getInputLeftExtensionSize();
    uint8_t getInputRightExtensionSize();
    regex024_error_code getErrno();
    bool haveSurvivors();
    bool isMatched();
    std::vector<REGEX_IS024_CAEvent> getMatchedThreadCABranchReverse();
    uint64_t getMatchedThreadSAValue(uint16_t key);

    regex024_error_code addNewMatchingThread();
    regex024_error_code extendedFeedCharacter(uint64_t input);
    regex024_error_code feedCharacter(uint64_t input, uint64_t bytesResembled);

private:
    bool gave_SOF = false;
    void* opaque;
};

#endif //LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H
