#ifndef LIBREGEXIS024_VM_OPCODES_H
#define LIBREGEXIS024_VM_OPCODES_H

#include <libregexis024vm/vm_opcodes_types.h>

namespace regex024_opcodes {
    enum regex024_opcode_I: uint8_t{
        /* READ <Settlement ID> */
        READ = 0,
        /* READZ = READ 0 */
        READZ = 1,
        /* JUMP <Near pointer> */
        JUMP = 2,

        /* JCEQUAL - jump conditional (equal): JCEQUAL <s1> <Near pointer> */
        JCEQUAL_B = 3,
        JCEQUAL_W = 4,
        JCEQUAL_DW = 5,
        JCEQUAL_QW = 6,
        /* JCLESS - jump conditional (less): JCLESS <s1> <Near pointer> */
        JCLESS_B = 7,
        JCLESS_W = 8,
        JCLESS_DW = 9,
        JCLESS_QW = 10,
        /* JCGRTR - jump conditional (greater): JCGRTR <s1> <Near pointer> */
        JCGRTR_B = 11,
        JCGRTR_W = 12,
        JCGRTR_DW = 13,
        JCGRTR_QW = 14,

        /* FORK <Settlemnt ID> <Near pointer> */
        FORK = 15,
        /* MATCH | */
        MATCH = 16,
        /* DIE | */
        DIE = 17,
        /* PARAM_READ_SS_NUMBER <Settlement ID (length)> */
        PARAM_READ_SS_NUMBER = 18,
        /* PARAM_FORK_SS_NUMBER <Settlement ID (length)> */
        PARAM_FORK_SS_NUMBER = 19,
        /* PARAM_SELARR_LEN <Track array index (length)> */
        PARAM_SELARR_LEN = 20,
        /* PARAM_COLSIFTFUNC_SET <Near pointer> */
        PARAM_COLSIFTFUNC_SET = 21,
        /* PARAM_COLSIFTFUNC_WIPE */
        PARAM_COLSIFTFUNC_WIPE = 22,
        /* MSG_MULTISTART_ALLOWED <1B> */
        MSG_MULTISTART_ALLOWED = 23,
        /* MSG_FED_INPUT_EXTENDED <1B> <1B> <Settlement ID (length of suffix)> */
        MSG_FED_INPUT_EXTENDED = 24,
        /* DMOVRABXSELARR <Track array index> */
        DMOV_RABX_SELARR = 25,
        /* DDISTRABXSELARR <Track array index> <Track array index> */
        DDIST_RABX_SELARR = 26,
        /* SIFTPRIOR_MIN_RABX */
        SIFTPRIOR_MIN_RABX = 27,
        /* SIFTPRIOR_MAX_RABX */
        SIFTPRIOR_MAX_RABX = 28,
        /* SIFT_DONE */
        SIFT_DONE = 29,
        /* MOV_COLARR_IMM <Track array index> <8B> */
        MOV_COLARR_IMM = 30,
        /* MOV_COLARR_BTPOS <Track array index> */
        MOV_COLARR_BTPOS = 31,
        /* MOV_SELARR_IMM <Track array index> <8B> */
        MOV_SELARR_IMM = 32,
        /* MOV_SELARR_CHPOS <Track array index> */
        MOV_SELARR_CHPOS = 33,
        /* INIT */
        INIT = 34,
        /* THROW */
        THROW = 35,
        regex024_opcode_greaterMax = 36
    };
}

typedef regex024_opcodes::regex024_opcode_I regex024_opcode;

const char* regex024_opcode_tostr(regex024_opcode x);


constexpr uint64_t REGEX024_BYTECODE_INSTRUCTION_SZ = 1;
constexpr uint64_t REGEX024_BYTECODE_SSLOT_ID_SZ = 4;
constexpr uint64_t REGEX024_BYTECODE_TRACK_ARRAY_INDEX_ID_SZ = 2;
constexpr uint64_t REGEX024_BYTECODE_NEAR_POINTER_SZ = 8;

bool vmprog_check_inboundness(regex_near_ptr_t prgSize, regex_near_ptr_t IP, regex_near_ptr_t region);

uint8_t vmprog_extract_b(regex_near_ptr_t* IPptr,  const uint8_t* prg);
uint16_t vmprog_extract_w(regex_near_ptr_t* IPptr, const uint8_t* prg);
uint32_t vmprog_extract_dw(regex_near_ptr_t* IPptr, const uint8_t* prg);
uint64_t vmprog_extract_qw(regex_near_ptr_t* IPptr, const uint8_t* prg);

uint8_t vmprog_extract_instruction(regex_near_ptr_t* IPptr, const uint8_t* prg);
regex_sslot_id_t vmprog_extract_sslot_id(regex_near_ptr_t* IPptr, const uint8_t* prg);
regex_near_ptr_t vmprog_extract_near_pointer(regex_near_ptr_t* IPptr, const uint8_t* prg);
regex_tai_t vmprog_extrack_track_array_index(regex_near_ptr_t* IPptr, const uint8_t* prg);

#endif //LIBREGEXIS024_VM_OPCODES_H
